from flask import request, Response

from .models import Session


PAGE_SIZE = 20


def get_ordering(args, model):
    default_ordering = ''
    if hasattr(model, 'Meta'):
        if hasattr(model.Meta, 'ordering'):
            default_ordering = model.Meta.ordering

    ordering_field = args.get('ordering', default_ordering)
    ordering_dir = 'desc' if ordering_field.startswith('-') else 'asc'
    ordering_field = ordering_field.strip('-')

    return ordering_field, ordering_dir


def get_pagination(args):
    page = int(args.get('page', 1))
    page_size = int(args.get('page_size', PAGE_SIZE))
    offset = (page-1) * page_size

    return offset, page_size


def generic_create_model(model, params):
    entity = model(**params)

    with Session.begin() as session:
        session.add(entity)
        session.flush()
        resp = entity.to_dict()

    return resp


def generic_list_view(model, filter_fields={}, **kwargs):
    extra_params = kwargs.get('extra_params', {})
    if request.method == 'GET':
        args = request.args
        filter_args = {field: args.get(field) for field in filter_fields if args.get(field)}
        filter_args.update(extra_params)
        ordering_field, ordering_dir = get_ordering(args, model)
        offset, limit = get_pagination(args)

        with Session() as session:
            entities = session.query(model).filter_by(**filter_args)
            count = entities.count()
            if ordering_field:
                order_by = getattr(getattr(model, ordering_field), ordering_dir)()
                entities = entities.order_by(order_by)
            entities = entities.offset(offset).limit(limit).all()

        data = [entity.to_dict(False) for entity in entities]
        resp = {
            'results': data,
            'count': count
        }

        return resp
    elif request.method == 'POST':
        try:
            params = request.get_json()
            params.update(extra_params)
            resp = generic_create_model(model, params)
            return resp
        except:
            return Response(status=400)
    elif request.method == 'DELETE':
        args = request.args
        filter_args = {field: args.get(field) for field in filter_fields if args.get(field)}
        filter_args.update(extra_params)

        with Session.begin() as session:
            session.query(model).filter_by(**filter_args).delete()

        return Response(status=200)


def generic_detail_view(model, id):
    with Session() as session:
        entity = session.query(model).filter_by(id=id).first()
    if not entity:
        return Response(status=404)

    if request.method == 'GET':
        return entity.to_dict()
    elif request.method == 'DELETE':
        with Session.begin() as session:
            session.query(model).filter_by(id=id).delete()

        return Response(status=200)
    elif request.method == 'PUT':
        params = request.get_json()
        try:
            with Session.begin() as session:
                entity = session.query(model).filter_by(id=id).first()
                for attr, val in params.items():
                    setattr(entity, attr, val)
                resp = entity.to_dict()
            return resp
        except:
            return Response(status=400)



import os
import json
import uuid
from datetime import datetime, timedelta

import requests

from sqlalchemy import create_engine, Column, DateTime, String, TEXT, ForeignKey, JSON, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker

from .constants import DATETIME_FORMAT

engine = create_engine(os.getenv('DB_URL'), pool_recycle=28700)
Session = sessionmaker(bind=engine)
Base = declarative_base()


def generate_uuid():
    return str(uuid.uuid4()).replace('-', '')


def period_mapping(period):
    now = datetime.utcnow()

    if period == 'Last Hour':
        _end_date = now.replace(minute=0, second=0, microsecond=0)
        start_date = (_end_date - timedelta(hours=1)).strftime(DATETIME_FORMAT)
    elif period == 'Last 4 Hours':
        _end_date = now.replace(minute=0, second=0, microsecond=0)
        start_date = (_end_date - timedelta(hours=4)).strftime(DATETIME_FORMAT)
    elif period == 'Last 8 Hours':
        _end_date = now.replace(minute=0, second=0, microsecond=0)
        start_date = (_end_date - timedelta(hours=8)).strftime(DATETIME_FORMAT)
    elif period == 'Last 24 Hours':
        _end_date = now.replace(minute=0, second=0, microsecond=0)
        start_date = (_end_date - timedelta(hours=24)).strftime(DATETIME_FORMAT)
    elif period == 'Last 7 Days':
        _end_date = now.replace(hour=0, minute=0, second=0, microsecond=0)
        start_date = (_end_date - timedelta(days=7)).strftime(DATETIME_FORMAT)
    elif period == 'Last 30 Days':
        _end_date = now.replace(hour=0, minute=0, second=0, microsecond=0)
        start_date = (_end_date - timedelta(days=30)).strftime(DATETIME_FORMAT)
    else:
        dates = period.split(' - ')
        start_date = datetime.strptime(dates[0], '%m/%d/%y').strftime(DATETIME_FORMAT)
        _end_date = datetime.strptime(dates[1], '%m/%d/%y')

    end_date = _end_date.strftime(DATETIME_FORMAT)

    return start_date, end_date


class DictMixin:
    def to_dict(self, detail_mode=None):
        raw_dict = self.__dict__
        result = {}
        for field in self.Meta.fields:
            val = raw_dict.get(field)
            if val is not None and isinstance(val, datetime):
                result[field] = val.strftime(DATETIME_FORMAT)
            else:
                result[field] = val

        return result


class InstalledApp(Base):
    __tablename__ = 'core_installedapp'

    id = Column(String(32), primary_key=True, default=generate_uuid)
    name = Column(String(250), nullable=False)
    slug = Column(String(100), nullable=False)
    version = Column(String(250), nullable=False)
    available_app_id = Column(String(250), nullable=False)
    app_domain = Column(String(200), nullable=False)
    created = Column(DateTime, default=datetime.utcnow)
    modified = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)

    analyses = relationship("Analysis", back_populates="app")


class Analysis(Base, DictMixin):
    __tablename__ = 'core_analysis'

    id = Column(String(32), primary_key=True, default=generate_uuid)
    app_id = Column(String(32), ForeignKey('core_installedapp.id'), nullable=False)
    name = Column(String(250), nullable=False)
    params = Column(JSON, nullable=False)
    created = Column(DateTime, default=datetime.utcnow)
    modified = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)
    tenant_id = Column(String(32))

    app = relationship("InstalledApp", back_populates="analyses", lazy=False)
    runs = relationship("AnalysisRun", back_populates="analysis", passive_deletes=True)
    sends = relationship("AnalysisSend", back_populates="analysis", passive_deletes=True)

    class Meta:
        fields = ('id', 'name', 'params', 'created', 'tenant_id')
        ordering = 'name'

    def run(self, func_compute, user_id, is_saved=False):
        from .utilities import upload_to_s3

        start_date, end_date = period_mapping(self.params['period'])
        result = func_compute(start_date, end_date)

        app_slug = 'anonymous'  # get the app slug
        if self.id:
            with Session() as session:
                analysis = session.query(Analysis).filter_by(id=self.id).first()
            app_slug = analysis.app.slug

        key = f'{app_slug}/run/{generate_uuid()}.json'
        result_url = upload_to_s3(json.dumps(result).encode(), key)

        run = AnalysisRun(
            params={
                'start_date': start_date,
                'end_date': end_date
            },
            analysis_id=self.id,
            triggered_by_id=user_id,
            is_saved=is_saved,
            result_url=result_url,
            date_completed=datetime.utcnow()
        )

        with Session.begin() as session:
            session.add(run)
            session.flush()
            run_id = run.id

        with Session() as session:
            run = session.query(AnalysisRun).filter_by(id=run_id).first()

        return run


class AnalysisRun(Base, DictMixin):
    __tablename__ = 'core_analysisrun'

    id = Column(String(32), primary_key=True, default=generate_uuid)
    analysis_id = Column(String(32), ForeignKey('core_analysis.id', ondelete='CASCADE'))
    params = Column(JSON)
    date_launched = Column(DateTime, default=datetime.utcnow)
    date_completed = Column(DateTime)
    result = Column(JSON)
    result_url = Column(String(300))
    is_saved = Column(Boolean, default=False)
    triggered_by_id = Column(String(32), nullable=False)  # user id

    analysis = relationship("Analysis", back_populates="runs", lazy=False)
    exports = relationship("AnalysisExport", back_populates="run", passive_deletes=True)

    class Meta:
        fields = ('id', 'analysis_id', 'params', 'date_launched', 'date_completed', 'is_saved')
        ordering = '-date_launched'

    def to_dict(self, detail_mode=True):
        resp = super().to_dict()
        resp['analysis_name'] = self.analysis.name

        return resp

    def get_result(self):
        result = {}
        if self.result_url:
            result = requests.get(self.result_url).json()
        elif self.result:  # legacy
            result = self.result

        return result


class AnalysisSend(Base, DictMixin):
    __tablename__ = 'core_analysissend'

    id = Column(String(32), primary_key=True, default=generate_uuid)
    analysis_id = Column(String(32), ForeignKey('core_analysis.id', ondelete='CASCADE'))
    subject = Column(String(200), nullable=False)
    format = Column(String(20), default='pdf')
    schedule = Column(String(50))
    recepients = Column(TEXT)
    message = Column(TEXT)
    attachment = Column(String(500))
    jwt_token = Column(TEXT)
    is_active = Column(Boolean, default=True)
    is_ran = Column(Boolean, default=False)
    last_ran = Column(DateTime, default=datetime.utcnow)
    created = Column(DateTime, default=datetime.utcnow)
    modified = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)

    analysis = relationship("Analysis", back_populates="sends", lazy=False)

    class Meta:
        fields = ('id', 'analysis_id', 'subject', 'format', 'schedule', 'recepients', 'message',
                  'attachment', 'is_active', 'is_ran', 'last_ran', 'created', 'modified')


class AnalysisExport(Base, DictMixin):
    __tablename__ = 'core_analysisexport'

    id = Column(String(32), primary_key=True, default=generate_uuid)
    run_id = Column(String(32), ForeignKey('core_analysisrun.id', ondelete='CASCADE'))
    url = Column(String(300))
    created = Column(DateTime, default=datetime.utcnow)
    modified = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)

    run = relationship("AnalysisRun", back_populates="exports", lazy=False)

    class Meta:
        fields = ('id', 'run_id', 'url', 'created', 'modified')



from flask import request

from .mixins import generic_list_view, generic_detail_view
from .models import Session, Analysis, AnalysisRun, AnalysisExport, AnalysisSend
from .utilities import (
    get_jwt_token, run_scheduled_send, generate_pdf, login_required,
    get_msp_id, get_user_id, call_get_requests, BASE_API_URL
)


#@login_required
def analysis_list_view():
    filter_fields = ['app_id']
    extra_params = {'tenant_id': get_msp_id()}
    return generic_list_view(Analysis, filter_fields, extra_params=extra_params)


#@login_required
def analysis_detail_view(id):
    return generic_detail_view(Analysis, id)


#@login_required
def analysis_run_list_view():
    filter_fields = ['analysis_id']
    extra_params = {'triggered_by_id': get_user_id()}
    return generic_list_view(AnalysisRun, filter_fields, extra_params=extra_params)


#@login_required
def analysis_run_detail_view(id):
    return generic_detail_view(AnalysisRun, id)


#@login_required
def analysis_export_create_view():
    resp = generic_list_view(AnalysisExport)

    with Session() as session:
        entity = session.query(AnalysisExport).filter_by(id=resp['id']).first()

    pdf_info = generate_pdf(entity.run)
    with Session.begin() as session:
        entity.url = pdf_info['Location']
        resp['url'] = pdf_info['Location']

    return resp


#@login_required
def analysis_export_detail_view(id):
    return generic_detail_view(AnalysisExport, id)


#@login_required
def analysis_send_list_view(func_compute):
    filter_fields = ['analysis_id']
    extra_params = {'jwt_token': get_jwt_token()}
    resp = generic_list_view(AnalysisSend, filter_fields, extra_params=extra_params)

    if request.method == 'POST':
        with Session() as session:
            entity = session.query(AnalysisSend).filter_by(id=resp['id']).first()

        if entity.is_active and not entity.is_ran and not entity.schedule:  # send now
            run_scheduled_send(entity, func_compute)

    return resp


#@login_required
def compute_view(func_compute):
    params = request.get_json()
    analysis = Analysis(
        id=params.get('analysis_id'),
        params=params
    )
    user_id = get_user_id()
    run = analysis.run(func_compute, user_id)

    resp = {
        'analysis': analysis.id,
        'analysis-run': run.id
    }

    return resp


#@login_required
def oap_get_users():
    org_id = get_msp_id()
    url = f'{BASE_API_URL}/api/v2/tenants/{org_id}/users/minimal'
    resp = call_get_requests(url, verify=False).json()

    return {
        'results': resp,
        'count': len(resp)
    }



# import random
# import dateutil.parser
# from datetime import datetime
# import logging
#
# from openpyxl import Workbook
# from openpyxl.styles import PatternFill, Font, Alignment
# from openpyxl.worksheet.table import Table, TableStyleInfo
# from openpyxl.utils import get_column_letter
# from openpyxl.chart import (
#     DoughnutChart,
#     ScatterChart,
#     BarChart,
#     Reference,
#     Series,
# )
# from openpyxl.chart.series import DataPoint
#
#
# class ExcelRenderer:
#     ROW_START = 3
#     logging.error(f'{ROW_START} <******  enter into ExcelRenderer /render/excel.py  ******>')
#     def __init__(self, analysis_run):
#         """
#         :param analysis_run: AnalysisRun
#         """
#         self.wb = Workbook()
#         self.analysis_run = analysis_run
#
#     def _load_data(self, ws, data, row_start, col_start, adjust_column=True, x_axis_date_format=None):
#         column_widths = {}
#         for row_delta, row in enumerate(data):
#             for col_delta, val in enumerate(row):
#                 col = col_start + col_delta
#                 cell = ws.cell(row=row_start+row_delta, column=col)
#                 if x_axis_date_format and row_delta > 0 and col_delta == 0:
#                     cell.value = datetime.strptime(val, x_axis_date_format).date()
#                 else:
#                     cell.value = val
#
#                 if col in column_widths:
#                     column_widths[col] = max(len(str(val)), column_widths[col])
#                 else:
#                     column_widths[col] = max(len(str(val)), ws.column_dimensions[get_column_letter(col)].width)
#
#         # handle cell width
#         if adjust_column:
#             for col, column_width in column_widths.items():
#                 ws.column_dimensions[get_column_letter(col)].width = column_width + 2
#
#     def add_table(self, ws, table_data):
#         """
#         add table component
#         """
#         row_start = ExcelRenderer.ROW_START + table_data['start-row'] + 1
#         col_start = table_data['start-col']
#         self._load_data(ws, table_data['data'], row_start, col_start)
#         row_span = len(table_data['data'])
#         col_span = len(table_data['data'][0])
#
#         ref = f'{get_column_letter(col_start)}{row_start}:{get_column_letter(col_start+col_span-1)}{row_start+row_span-1}'
#         table_name = f'Table{random.randint(0, 100)}'
#
#         table = Table(displayName=table_name, ref=ref, tableStyleInfo=TableStyleInfo(name="TableStyleMedium9"))
#         ws.add_table(table)
#
#     def add_doughnut_chart(self, ws, chart_data):
#         """
#         add doughnut chart component
#         """
#         row_start = ExcelRenderer.ROW_START + chart_data['start-row']
#         col_start = chart_data['start-col']
#         self._load_data(ws, chart_data['data'], row_start, col_start, False)
#         row_span = len(chart_data['data'])
#
#         chart = DoughnutChart(holeSize=50)
#         labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
#         data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
#         chart.add_data(data, titles_from_data=True)
#         chart.set_categories(labels)
#         chart.title = chart_data['chart-title']
#         chart.style = 12
#         chart.width = chart_data.get('width', 11)
#         chart.height = chart_data.get('height', 5)
#
#         slices = [DataPoint(idx=i) for i in range(row_span-1)]
#         chart.series[0].data_points = slices
#         colors = ["0077C8", "32a3df", "673ab7", "9c27b0"]
#
#         for idx, slice in enumerate(slices):
#             slice.graphicalProperties.solidFill = colors[idx % 4]
#
#         ws.add_chart(chart, chart_data['chart-position'])
#
#     def add_bar_chart(self, ws, chart_data):
#         """
#         add bar chart component
#         """
#         row_start = ExcelRenderer.ROW_START + chart_data['start-row']
#         col_start = chart_data['start-col']
#         self._load_data(ws, chart_data['data'], row_start, col_start, False)
#         row_span = len(chart_data['data'])
#
#         chart = BarChart()
#         chart.type = "col"
#         chart.style = 10
#         chart.title = chart_data['chart-title']
#         chart.x_axis.title = chart_data.get('x-axis-title')
#         chart.y_axis.title = chart_data.get('y-axis-title')
#
#         labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
#         data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
#         chart.add_data(data, titles_from_data=True)
#         chart.set_categories(labels)
#         chart.style = 4
#         chart.legend = None
#         chart.width = chart_data.get('width', 17)
#         chart.height = chart_data.get('height', 8.5)
#
#         s1 = chart.series[0]
#         s1.graphicalProperties.line.solidFill = "0077C8"
#         s1.graphicalProperties.solidFill = "0077C8"
#
#         ws.add_chart(chart, chart_data['chart-position'])
#
#     def add_scatter_chart(self, ws, chart_data):
#         """
#         add scatter chart component
#         """
#         row_start = ExcelRenderer.ROW_START + chart_data['start-row']
#         col_start = chart_data['start-col']
#         x_axis_date_format = chart_data.get('x-axis-date-format')
#         self._load_data(ws, chart_data['data'], row_start, col_start, False, x_axis_date_format)
#         row_span = len(chart_data['data'])
#
#         chart = ScatterChart()
#         chart.style = 5
#         chart.legend = None
#         chart.title = chart_data['chart-title']
#         chart.x_axis.title = chart_data.get('x-axis-title')
#         chart.y_axis.title = chart_data.get('y-axis-title')
#         chart.width = chart_data.get('width', 15)
#         chart.height = chart_data.get('height', 7.5)
#
#         labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
#         data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
#         series = Series(data, labels, title_from_data=True)
#         chart.series.append(series)
#
#         s1 = chart.series[0]
#         s1.marker.symbol = "circle"
#         s1.marker.graphicalProperties.solidFill = "0077C8"  # Marker filling
#         s1.marker.graphicalProperties.line.solidFill = "0077C8"  # Marker outline
#         s1.graphicalProperties.line.solidFill = "0077C8"
#         s1.graphicalProperties.line.width = 24050  # width in EMUs
#         if x_axis_date_format:
#             chart.x_axis.number_format = 'm/d'
#             chart.x_axis.majorTimeUnit = "days"
#
#         ws.add_chart(chart, chart_data['chart-position'])
#
#     def generate_summary(self):
#         """
#         add summary sheet
#         """
#         ws = self.wb.active
#         ws.title = 'Summary'
#
#         for row_idx in range(1, 30):
#             row = ws.row_dimensions[row_idx]
#             row.fill = PatternFill("solid", fgColor="eeeeee")
#
#         ws.column_dimensions['B'].width = 24
#         ws.column_dimensions['C'].width = 24
#
#         ws.merge_cells('B3:Z4')
#         cell = ws['B3']
#         cell.value = self.analysis_run.analysis.app.name
#         cell.font = Font(color="00598B", bold=True)
#         cell.alignment = Alignment(vertical="center")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['B5']
#         cell.value = 'Analysis Parameters'
#         cell.font = Font(color="00598B")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['B6']
#         cell.value = 'Client'
#         cell.alignment = Alignment(horizontal="right")
#         cell.font = Font(color="00598B")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['C6']
#         cell.value = self.analysis_run.analysis.name
#         cell.font = Font(color="000000")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['B7']
#         cell.value = 'Analysis Period'
#         cell.alignment = Alignment(horizontal="right")
#         cell.font = Font(color="00598B")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['C7']
#         params = self.analysis_run.params
#         start_date = dateutil.parser.isoparse(params["start_date"])
#         end_date = dateutil.parser.isoparse(params["end_date"])
#         cell.value = f'{start_date.strftime("%B %d, %Y")} - {end_date.strftime("%B %d, %Y")}'
#         cell.font = Font(color="000000")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['B8']
#         cell.value = 'Analysis'
#         cell.font = Font(color="00598B")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['B9']
#         cell.value = 'Analysis Name'
#         cell.alignment = Alignment(horizontal="right")
#         cell.font = Font(color="00598B")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['C9']
#         cell.value = self.analysis_run.analysis.name
#         cell.font = Font(color="000000")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['B10']
#         cell.value = 'Run date'
#         cell.alignment = Alignment(horizontal="right")
#         cell.font = Font(color="00598B")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['C10']
#         cell.value = self.analysis_run.date_launched.strftime('%H:%M %p %Z, %B %d, %Y')
#         cell.font = Font(color="000000")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['B11']
#         cell.value = 'Complete date'
#         cell.alignment = Alignment(horizontal="right")
#         cell.font = Font(color="00598B")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['C11']
#         cell.value = self.analysis_run.date_completed.strftime('%H:%M %p %Z, %B %d, %Y')
#         cell.font = Font(color="000000")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['B12']
#         cell.value = 'User'
#         cell.alignment = Alignment(horizontal="right")
#         cell.font = Font(color="00598B")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         # cell = ws['C12']
#         # user = self.analysis_run.analysis.app.installed_by
#         # cell.value = f'{user.first_name} {user.last_name}'
#         # cell.font = Font(color="000000")
#         # cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['B13']
#         cell.value = 'App version'
#         cell.alignment = Alignment(horizontal="right")
#         cell.font = Font(color="00598B")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#         cell = ws['C13']
#         cell.value = self.analysis_run.analysis.app.version
#         cell.font = Font(color="000000")
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#
#     def add_header(self, ws, sheet_data):
#         """
#         add header to the sheet
#         """
#         for row_idx in range(1, ExcelRenderer.ROW_START):
#             row = ws.row_dimensions[row_idx]
#             row.fill = PatternFill("solid", fgColor="eeeeee")
#
#         ws.merge_cells('B1:Z2')
#         cell = ws['B1']
#         cell.value = self.analysis_run.analysis.app.name
#         cell.font = Font(color="00598B", bold=True)
#         cell.fill = PatternFill("solid", fgColor="eeeeee")
#         cell.alignment = Alignment(vertical="center")
#
#     def add_component_title(self, ws, component_data):
#         """
#         add component title
#         """
#         cell = ws.cell(row=ExcelRenderer.ROW_START+component_data['start-row'], column=component_data['start-col'])
#         cell.value = component_data['title']
#         color = component_data.get('color', '00598B')
#         cell.font = Font(color=color, bold=True)
#
#     def add_component(self, ws, component_data):
#         if component_data['type'] == 'table':
#             if component_data.get('title'):
#                 self.add_component_title(ws, component_data)
#             self.add_table(ws, component_data)
#         elif component_data['type'] == 'doughnut-chart':
#             self.add_doughnut_chart(ws, component_data)
#         elif component_data['type'] == 'bar-chart':
#             self.add_bar_chart(ws, component_data)
#         elif component_data['type'] == 'scatter-chart':
#             self.add_scatter_chart(ws, component_data)
#
#     def render(self):
#         """
#         :return: Workbook
#         """
#         self.generate_summary()
#         run_data = self.analysis_run.get_result().get('excel-data', {})
#
#         for sheet in run_data['sheets']:
#             ws = self.wb.create_sheet(title=sheet['title'])
#             self.add_header(ws, sheet)
#
#             for component_data in sheet['components']:
#                 self.add_component(ws, component_data)
#
#         return self.wb


#==========================    Modified Code    ======================================
import random
import dateutil.parser
from datetime import datetime
import logging

from openpyxl import Workbook
from openpyxl.styles import PatternFill, Font, Alignment
from openpyxl.worksheet.table import Table, TableStyleInfo
from openpyxl.utils import get_column_letter
from openpyxl.chart import (
    DoughnutChart,
    ScatterChart,
    BarChart,
    Reference,
    Series,
)
from openpyxl.chart.series import DataPoint


class ExcelRenderer:
    ROW_START = 3
    logging.error(f'{ROW_START} <-----   Enter Into Excel Process   ----->')
    def __init__(self, analysis_run):
        """
        :param analysis_run: AnalysisRun
        """
        self.wb = Workbook()
        self.analysis_run = analysis_run

    def _load_data(self, ws, data, row_start, col_start, adjust_column=True, x_axis_date_format=None):
        column_widths = {}
        for row_delta, row in enumerate(data):
            for col_delta, val in enumerate(row):
                col = col_start + col_delta
                cell = ws.cell(row=row_start+row_delta, column=col)
                if x_axis_date_format and row_delta > 0 and col_delta == 0:
                    cell.value = datetime.strptime(val, x_axis_date_format).date()
                else:
                    cell.value = val

                if col in column_widths:
                    column_widths[col] = max(len(str(val)), column_widths[col])
                else:
                    column_widths[col] = max(len(str(val)), ws.column_dimensions[get_column_letter(col)].width)

        # handle cell width
        if adjust_column:
            for col, column_width in column_widths.items():
                ws.column_dimensions[get_column_letter(col)].width = column_width + 2

    def add_table(self, ws, table_data):
        """
        add table component
        """
        row_start = ExcelRenderer.ROW_START + table_data['start-row'] + 1
        col_start = table_data['start-col']
        self._load_data(ws, table_data['data'], row_start, col_start)
        row_span = len(table_data['data'])
        col_span = len(table_data['data'][0])

        ref = f'{get_column_letter(col_start)}{row_start}:{get_column_letter(col_start+col_span-1)}{row_start+row_span-1}'
        table_name = f'Table{random.randint(0, 100)}'

        table = Table(displayName=table_name, ref=ref, tableStyleInfo=TableStyleInfo(name="TableStyleMedium9"))
        ws.add_table(table)

    def add_doughnut_chart(self, ws, chart_data):
        """
        add doughnut chart component
        """
        row_start = ExcelRenderer.ROW_START + chart_data['start-row']
        col_start = chart_data['start-col']
        self._load_data(ws, chart_data['data'], row_start, col_start, False)
        row_span = len(chart_data['data'])

        chart = DoughnutChart(holeSize=50)
        labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
        data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
        chart.add_data(data, titles_from_data=True)
        chart.set_categories(labels)
        chart.title = chart_data['chart-title']
        chart.style = 12
        chart.width = chart_data.get('width', 11)
        chart.height = chart_data.get('height', 5)

        slices = [DataPoint(idx=i) for i in range(row_span-1)]
        chart.series[0].data_points = slices
        colors = ["0077C8", "32a3df", "673ab7", "9c27b0"]

        for idx, slice in enumerate(slices):
            slice.graphicalProperties.solidFill = colors[idx % 4]

        ws.add_chart(chart, chart_data['chart-position'])

    def add_bar_chart(self, ws, chart_data):
        """
        add bar chart component
        """
        row_start = ExcelRenderer.ROW_START + chart_data['start-row']
        col_start = chart_data['start-col']
        self._load_data(ws, chart_data['data'], row_start, col_start, False)
        row_span = len(chart_data['data'])

        chart = BarChart()
        chart.type = "col"
        chart.style = 10
        chart.title = chart_data['chart-title']
        chart.x_axis.title = chart_data.get('x-axis-title')
        chart.y_axis.title = chart_data.get('y-axis-title')

        labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
        data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
        chart.add_data(data, titles_from_data=True)
        chart.set_categories(labels)
        chart.style = 4
        chart.legend = None
        chart.width = chart_data.get('width', 17)
        chart.height = chart_data.get('height', 8.5)

        s1 = chart.series[0]
        s1.graphicalProperties.line.solidFill = "0077C8"
        s1.graphicalProperties.solidFill = "0077C8"

        ws.add_chart(chart, chart_data['chart-position'])

    def add_scatter_chart(self, ws, chart_data):
        """
        add scatter chart component
        """
        row_start = ExcelRenderer.ROW_START + chart_data['start-row']
        col_start = chart_data['start-col']
        x_axis_date_format = chart_data.get('x-axis-date-format')
        self._load_data(ws, chart_data['data'], row_start, col_start, False, x_axis_date_format)
        row_span = len(chart_data['data'])

        chart = ScatterChart()
        chart.style = 5
        chart.legend = None
        chart.title = chart_data['chart-title']
        chart.x_axis.title = chart_data.get('x-axis-title')
        chart.y_axis.title = chart_data.get('y-axis-title')
        chart.width = chart_data.get('width', 15)
        chart.height = chart_data.get('height', 7.5)

        labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
        data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
        series = Series(data, labels, title_from_data=True)
        chart.series.append(series)

        s1 = chart.series[0]
        s1.marker.symbol = "circle"
        s1.marker.graphicalProperties.solidFill = "0077C8"  # Marker filling
        s1.marker.graphicalProperties.line.solidFill = "0077C8"  # Marker outline
        s1.graphicalProperties.line.solidFill = "0077C8"
        s1.graphicalProperties.line.width = 24050  # width in EMUs
        if x_axis_date_format:
            chart.x_axis.number_format = 'm/d'
            chart.x_axis.majorTimeUnit = "days"

        ws.add_chart(chart, chart_data['chart-position'])

    def generate_summary(self):
        """
        add summary sheet
        """
        ws = self.wb.active
        ws.title = 'Summary'

        for row_idx in range(1, 30):
            row = ws.row_dimensions[row_idx]
            row.fill = PatternFill("solid", fgColor="eeeeee")

        ws.column_dimensions['B'].width = 24
        ws.column_dimensions['C'].width = 24

        ws.merge_cells('B3:Z4')
        cell = ws['B3']
        cell.value = "asset-insights"
        cell.font = Font(color="00598B", bold=True)
        cell.alignment = Alignment(vertical="center")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['B5']
        cell.value = 'Analysis Parameters'
        cell.font = Font(color="00598B")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['B6']
        cell.value = 'Client'
        cell.alignment = Alignment(horizontal="right")
        cell.font = Font(color="00598B")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['C6']
        cell.value = "Dec-11-analysis"
        cell.font = Font(color="000000")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['B7']
        cell.value = 'Analysis Period'
        cell.alignment = Alignment(horizontal="right")
        cell.font = Font(color="00598B")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['C7']
        # params = self.analysis_run.params
        start_date = "2021-12-10T12:00:00"
        end_date = "2021-12-11T12:00:00"
        cell.value = f'{start_date} - {end_date}'
        cell.font = Font(color="000000")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['B8']
        cell.value = 'Analysis'
        cell.font = Font(color="00598B")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['B9']
        cell.value = 'Analysis Name'
        cell.alignment = Alignment(horizontal="right")
        cell.font = Font(color="00598B")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['C9']
        cell.value = "Dec analysis"
        cell.font = Font(color="000000")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['B10']
        cell.value = 'Run date'
        cell.alignment = Alignment(horizontal="right")
        cell.font = Font(color="00598B")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['C10']
        cell.value = "2021-12-11T12:00:00"
        cell.font = Font(color="000000")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['B11']
        cell.value = 'Complete date'
        cell.alignment = Alignment(horizontal="right")
        cell.font = Font(color="00598B")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['C11']
        cell.value = "2021-12-12T12:00:00"
        cell.font = Font(color="000000")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['B12']
        cell.value = 'User'
        cell.alignment = Alignment(horizontal="right")
        cell.font = Font(color="00598B")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        # cell = ws['C12']
        # user = self.analysis_run.analysis.app.installed_by
        # cell.value = f'{user.first_name} {user.last_name}'
        # cell.font = Font(color="000000")
        # cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['B13']
        cell.value = 'App version'
        cell.alignment = Alignment(horizontal="right")
        cell.font = Font(color="00598B")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

        cell = ws['C13']
        cell.value = "11.O"
        cell.font = Font(color="000000")
        cell.fill = PatternFill("solid", fgColor="eeeeee")

    def add_header(self, ws, sheet_data):
        """
        add header to the sheet
        """
        for row_idx in range(1, ExcelRenderer.ROW_START):
            row = ws.row_dimensions[row_idx]
            row.fill = PatternFill("solid", fgColor="eeeeee")

        ws.merge_cells('B1:Z2')
        cell = ws['B1']
        cell.value = "Asset-Insights"
        cell.font = Font(color="00598B", bold=True)
        cell.fill = PatternFill("solid", fgColor="eeeeee")
        cell.alignment = Alignment(vertical="center")

    def add_component_title(self, ws, component_data):
        """
        add component title
        """
        cell = ws.cell(row=ExcelRenderer.ROW_START+component_data['start-row'], column=component_data['start-col'])
        cell.value = component_data['title']
        color = component_data.get('color', '00598B')
        cell.font = Font(color=color, bold=True)

    def add_component(self, ws, component_data):
        if component_data['type'] == 'table':
            if component_data.get('title'):
                self.add_component_title(ws, component_data)
            self.add_table(ws, component_data)
        elif component_data['type'] == 'doughnut-chart':
            self.add_doughnut_chart(ws, component_data)
        elif component_data['type'] == 'bar-chart':
            self.add_bar_chart(ws, component_data)
        elif component_data['type'] == 'scatter-chart':
            self.add_scatter_chart(ws, component_data)

    def render(self):
        """
        :return: Workbook
        """
        self.generate_summary()
        # run_data = self.analysis_run.get_result().get('excel-data', {})
        run_data = self.analysis_run
        for sheet in run_data['sheets']:
            ws = self.wb.create_sheet(title=sheet['title'])
            self.add_header(ws, sheet)

            for component_data in sheet['components']:
                self.add_component(ws, component_data)

        return self.wb







import os
import time
import logging
from io import BytesIO
from datetime import datetime

import json

from urllib.request import urlopen
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart

from croniter import croniter
from flask.wrappers import Response

import jwt
import boto3
import flask
import requests

from .renderer.excel import ExcelRenderer
from .models import Session, AnalysisRun, AnalysisSend
from .constants import DATETIME_FORMAT


BASE_API_URL = os.getenv('DATA_API_BASE_URL', '')
AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID')
AWS_SECRET_ACCESS_KEY = os.getenv('AWS_SECRET_ACCESS_KEY')
REGION_NAME = os.getenv('REGION_NAME')
S3_BUCKET_NAME = os.getenv('S3_BUCKET_NAME')
API_RETRY = int(os.getenv('API_RETRY', 3))
API_KEY = os.getenv('API_KEY')
API_SECRET = os.getenv('API_SECRET')
CLIENT_TOKEN = 'msp_28'
ACCESS_HEADER = None

def get_token():
    headers = {"Content-Type" : "application/x-www-form-urlencoded" , "Accept" : "application/json"};
    post_data = {"grant_type": "client_credentials", "client_id" : API_KEY, "client_secret" : API_SECRET};
    token_url = BASE_API_URL + "/auth/oauth/token";
    response = requests.post(token_url,data=post_data,headers=headers,verify=False);
    json = response.json();
    auth = str(json["token_type"]) + " " + str(json["access_token"])
    return auth


def get_jwt_token():
    try:
        jwt_token = flask.request.cookies.get('OPSRAMP_JWT_TOKEN', '')
    except:  # outside Flask
        jwt_token = os.getenv('OPSRAMP_JWT_TOKEN', '')

    return jwt_token


def get_headers():
    headers = {
        'Authorization': f'Bearer {get_jwt_token()}'
    }

    return headers


def get_msp_id():
    msp_id = None
    jwt_token = get_jwt_token()
    if jwt_token:
        decoded = jwt.decode(jwt_token, options={"verify_signature": False})
        msp_id = decoded['orgId']

    return msp_id


def get_user_id():
    user_id = None
    jwt_token = get_jwt_token()
    if jwt_token:
        decoded = jwt.decode(jwt_token, options={"verify_signature": False})
        user_id = decoded['userId']

    return user_id


def call_requests(method, url, params=None, data=None, json=None, verify=True):
    headers = get_headers()
    retry = 1
    while retry <= API_RETRY:
        try:
            resp = requests.request(method, url, params=params, data=data, json=json, headers=headers, verify=verify)
        except requests.exceptions.ConnectionError:
            time.sleep(retry * 2)
            continue

        return resp


def call_get_requests(url, params=None, verify=True):
    return call_requests('GET', url, params, verify=verify)


def call_post_requests(url, params=None, data=None, verify=True):
    return call_requests('POST', url, params, data, verify=verify)


def is_authenticated():
    REQUIRE_AUTH_REDIRECT = os.getenv('REQUIRE_AUTH_REDIRECT') == 'true'
    if not REQUIRE_AUTH_REDIRECT:
        return True

    url = f'{BASE_API_URL}/api/v2/users/me'
    res = call_get_requests(url)

    return res.status_code == 200


def login_required(view):
    '''Decorator that check authentication'''
  
    def wrap(*args, **kwargs):
        if not is_authenticated():
            return Response('Not authorized', status=401)
        result = view(*args, **kwargs)
        return result
    return wrap


def get_epoc_from_datetime_string(str_datetime):
    timestamp = datetime.strptime(str_datetime, DATETIME_FORMAT).timestamp()
    return timestamp

def get_run_result(run_id, field=None, default_value=None):
    with Session() as session:
        run = session.query(AnalysisRun).filter_by(id=run_id).first()

    result = run.get_result() if run else {}
    if field:
        result = result.get(field, default_value)

    return result

def get_result_by_run(run_id, field=None, default_value=None):
    logging.error(f'{" "} <-----   Entered Into get result by run   ----->')
    s3 = get_s3_client()
    run_id= f'analytics-apps/json/{run_id}'
    res_object = s3.get_object(Bucket=S3_BUCKET_NAME,Key=run_id)
    serializedObject = res_object['Body'].read()
    result = json.loads(serializedObject)
    if field:
        result = result.get(field, default_value)
    return result


def get_run_params(run_id, field=None, default_value=None):
    with Session() as session:
        run = session.query(AnalysisRun).filter_by(id=run_id).first()

    if run:
        return run.params


def get_ses_client():
    return boto3.client('ses',
                        region_name=REGION_NAME,
                        aws_access_key_id=AWS_ACCESS_KEY_ID,
                        aws_secret_access_key=AWS_SECRET_ACCESS_KEY)


def get_s3_client():
    return boto3.client('s3',
                        region_name=REGION_NAME,
                        aws_access_key_id=AWS_ACCESS_KEY_ID,
                        aws_secret_access_key=AWS_SECRET_ACCESS_KEY)
    

def send_email(subject, from_email, to_emails, body, attachment=None):
    message = MIMEMultipart()
    message['Subject'] = subject
    message['From'] = from_email
    message['To'] = to_emails

    # message body
    part = MIMEText(body, 'html')
    message.attach(part)

    if attachment:
        attachment_body = urlopen(attachment).read()
        part = MIMEApplication(attachment_body)
        part.add_header('Content-Disposition', 'attachment', filename=attachment)
        message.attach(part)

    resp = get_ses_client().send_raw_email(
        Source=message['From'],
        Destinations=to_emails.split(','),
        RawMessage={
            'Data': message.as_string()
        }
    )

    return resp


def upload_to_s3(content, location):
    '''
    :param: content: bytes
    :param: location: str
    '''
    s3 = boto3.resource('s3',
                        region_name=REGION_NAME,
                        aws_access_key_id=AWS_ACCESS_KEY_ID,
                        aws_secret_access_key=AWS_SECRET_ACCESS_KEY)
    object_url = f'https://{S3_BUCKET_NAME}.s3.{REGION_NAME}.amazonaws.com/{location}'

    try:
        s3.Bucket(S3_BUCKET_NAME).put_object(Body=content,
                                             Key=location)
        return object_url
    except Exception:
        pass
    

def generate_pdf(analysis_run):
    logging.error(f'{"Pdf"} <-----   Entered Into Pdf Generation Process   ----->')
    url = os.getenv("PDF_SERVICE")
    # analysis_run = json.dumps(analysis_run)
    data = {
        'domain': 'https://asura.opsramp.net',
        'report': 'analytics-apps',
        'run': analysis_run,
        'route': '/full-view',
        'token': get_token(),
        'size': 'A4'
    }

    res = requests.post(url, data=data).json()
    logging.error(f'{res} <-----     Pdf Response     ----->')
    return res



def generate_excel(analysis_run, file_name=None):
    try:
        excel_renderer = ExcelRenderer(analysis_run)
        workbook = excel_renderer.render()
    except Exception as ex:
        raise ex

    if file_name:
        output = None
        workbook.save(file_name)
    else:
        output = BytesIO()
        workbook.save(output)

    return output


def upload_excel(analysis_run, excel_file):
    timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    s3_path = f'{analysis_run.analysis.app.slug}/excel/{timestamp}.xlsx'

    return upload_to_s3(excel_file, s3_path)


def run_scheduled_send(analysis_send, func_compute):
    analysis_run = analysis_send.analysis.run(func_compute, None, True)

    if analysis_run:
        if analysis_send.format == 'pdf':
            pdf_info = generate_pdf(analysis_run)
            attachment = pdf_info['Location']
        elif analysis_send.format == 'excel':
            excel_file = generate_excel(analysis_run)
            if excel_file:
                attachment = upload_excel(analysis_run, excel_file.getvalue())
            else:
                attachment = None

        with Session.begin() as session:
            update_send = session.query(AnalysisSend).filter_by(id=analysis_send.id).first()
            update_send.attachment = attachment
            update_send.is_ran = True

        to_emails = analysis_send.recepients
        send_email(analysis_send.subject, os.getenv('FROM_EMAIL'), to_emails, analysis_send.message, attachment)


def check_schedule_send(func_compute):
    app_id = os.getenv('OAP_APP_ID')
    now = datetime.now()
    now_hour = datetime(now.year, now.month, now.day, now.hour)

    with Session() as session:
        sends = session.query(AnalysisSend) \
                       .filter_by(is_active=True) \
                       .filter(AnalysisSend.schedule.is_not(None)) \
                       .join(AnalysisSend.analysis) \
                       .filter_by(app_id=app_id).all()

    for send in sends:
        if croniter.is_valid(send.schedule) and croniter.match(send.schedule, now_hour):
            os.environ['OPSRAMP_JWT_TOKEN'] = send.jwt_token
            run_scheduled_send(send, func_compute)
            print(send.id, '='*10)
